@@ -102,7 +102,7 @@ size_t gnutls_record_check_pending(gnutls_session_t session)
  * @session: is a #gnutls_session_t structure.
  *
  * This function checks if there pending corked
- * data in the gnutls buffers --see gnutls_cork(). 
+ * data in the gnutls buffers --see gnutls_record_cork(). 
  *
  * Returns: Returns the size of the corked data or zero.
  *
@@ -160,6 +160,35 @@ _gnutls_record_buffer_get(content_type_t type,
 	return length;
 }
 
+int
+_gnutls_record_buffer_get_packet(content_type_t type, gnutls_session_t session, gnutls_packet_t *packet)
+{
+	mbuffer_st *bufel;
+
+	bufel =
+	    _mbuffer_head_pop_first(&session->internals.record_buffer);
+	if (bufel == NULL)
+		return
+		    gnutls_assert_val
+		    (GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE);
+
+	if (type != bufel->type) {
+		if (IS_DTLS(session))
+			_gnutls_audit_log(session,
+					  "Discarded unexpected %s (%d) packet (expecting: %s)\n",
+					  _gnutls_packet2str(bufel->type),
+					  (int) bufel->type,
+					  _gnutls_packet2str(type));
+		_mbuffer_head_remove_bytes(&session->internals.
+					   record_buffer, bufel->msg.size);
+		return gnutls_assert_val(GNUTLS_E_UNEXPECTED_PACKET);
+	}
+
+	*packet = bufel;
+
+	return bufel->msg.size - bufel->mark;
+}
+
 inline static void reset_errno(gnutls_session_t session)
 {
 	session->internals.errnum = 0;
@@ -221,7 +250,7 @@ _gnutls_dgram_read(gnutls_session_t session, mbuffer_st ** bufel,
 		gettime(&t1);
 	}
 
-	*bufel = _mbuffer_alloc(0, max_size);
+	*bufel = _mbuffer_alloc_align16(max_size, get_total_headers(session));
 	if (*bufel == NULL)
 		return gnutls_assert_val(GNUTLS_E_MEMORY_ERROR);
 
@@ -233,9 +262,8 @@ _gnutls_dgram_read(gnutls_session_t session, mbuffer_st ** bufel,
 	if (i < 0) {
 		int err = get_errno(session);
 
-		_gnutls_read_log
-		    ("READ: %d returned from %p, errno=%d gerrno=%d\n",
-		     (int) i, fd, errno, session->internals.errnum);
+		_gnutls_read_log("READ: %d returned from %p, errno=%d\n",
+		     		 (int) i, fd, err);
 
 		ret = errno_to_gerr(err, 1);
 		goto cleanup;
@@ -289,7 +317,7 @@ _gnutls_stream_read(gnutls_session_t session, mbuffer_st ** bufel,
 
 	session->internals.direction = 0;
 
-	*bufel = _mbuffer_alloc(0, MAX(max_size, size));
+	*bufel = _mbuffer_alloc_align16(MAX(max_size, size), get_total_headers(session));
 	if (!*bufel) {
 		gnutls_assert();
 		return GNUTLS_E_MEMORY_ERROR;
@@ -394,18 +422,25 @@ _gnutls_read(gnutls_session_t session, mbuffer_st ** bufel,
 					   ms);
 }
 
+/* @vec: if non-zero then the vector function will be used to
+ *       push the data.
+ */
 static ssize_t
 _gnutls_writev_emu(gnutls_session_t session, gnutls_transport_ptr_t fd,
-		   const giovec_t * giovec, unsigned int giovec_cnt)
+		   const giovec_t * giovec, unsigned int giovec_cnt, unsigned vec)
 {
 	unsigned int j = 0;
 	size_t total = 0;
 	ssize_t ret = 0;
 
 	for (j = 0; j < giovec_cnt; j++) {
-		ret =
-		    session->internals.push_func(fd, giovec[j].iov_base,
-						 giovec[j].iov_len);
+		if (vec) {
+			ret = session->internals.vec_push_func(fd, &giovec[j], 1);
+		} else {
+			ret =
+			    session->internals.push_func(fd, giovec[j].iov_base,
+							 giovec[j].iov_len);
+		}
 
 		if (ret == -1) {
 		        gnutls_assert();
@@ -424,27 +459,43 @@ _gnutls_writev_emu(gnutls_session_t session, gnutls_transport_ptr_t fd,
 	return ret;
 }
 
-
+/* @total: The sum of the data in giovec
+ */
 static ssize_t
 _gnutls_writev(gnutls_session_t session, const giovec_t * giovec,
-	       int giovec_cnt)
+	       unsigned giovec_cnt, unsigned total)
 {
 	int i;
+	bool is_dtls = IS_DTLS(session);
+	unsigned no_writev = 0;
 	gnutls_transport_ptr_t fd = session->internals.transport_send_ptr;
 
 	reset_errno(session);
 
-	if (session->internals.push_func != NULL)
-		i = _gnutls_writev_emu(session, fd, giovec, giovec_cnt);
-	else
-		i = session->internals.vec_push_func(fd, giovec,
-						     giovec_cnt);
+	if (session->internals.vec_push_func != NULL) {
+		if (is_dtls && giovec_cnt > 1) {
+			if (total > session->internals.dtls.mtu) {
+				no_writev = 1;
+			}
+		}
+
+		if (no_writev == 0) {
+			i = session->internals.vec_push_func(fd, giovec,
+							     giovec_cnt);
+		} else {
+			i = _gnutls_writev_emu(session, fd, giovec, giovec_cnt, 1);
+		}
+	} else if (session->internals.push_func != NULL) {
+		i = _gnutls_writev_emu(session, fd, giovec, giovec_cnt, 0);
+	} else
+		return gnutls_assert_val(GNUTLS_E_PUSH_ERROR);
 
 	if (i == -1) {
 		int err = get_errno(session);
-		_gnutls_debug_log("errno: %d\n", err);
+		_gnutls_debug_log("WRITE: %d returned from %p, errno: %d\n",
+				  i, fd, err);
 
-		return errno_to_gerr(err, IS_DTLS(session));
+		return errno_to_gerr(err, is_dtls);
 	}
 	return i;
 }
@@ -607,6 +658,7 @@ ssize_t _gnutls_io_write_flush(gnutls_session_t session)
 	int i = 0;
 	mbuffer_st *cur;
 
+	session->internals.direction = 1;
 	_gnutls_write_log("WRITE FLUSH: %d bytes in buffer.\n",
 			  (int) send_buffer->byte_length);
 
@@ -628,7 +680,7 @@ ssize_t _gnutls_io_write_flush(gnutls_session_t session)
 		return 0;
 	}
 
-	ret = _gnutls_writev(session, iovec, i);
+	ret = _gnutls_writev(session, iovec, i, tosend);
 	if (ret >= 0) {
 		_mbuffer_head_remove_bytes(send_buffer, ret);
 		_gnutls_write_log
@@ -924,14 +976,14 @@ static int merge_handshake_packet(gnutls_session_t session,
 		}
 	}
 
-	if (exists == 0)
+	if (!exists)
 		pos = session->internals.handshake_recv_buffer_size;
 
 	if (pos >= MAX_HANDSHAKE_MSGS)
 		return
 		    gnutls_assert_val(GNUTLS_E_TOO_MANY_HANDSHAKE_PACKETS);
 
-	if (exists == 0) {
+	if (!exists) {
 		if (hsk->length > 0 && hsk->end_offset > 0
 		    && hsk->end_offset - hsk->start_offset + 1 !=
 		    hsk->length) {
@@ -962,7 +1014,7 @@ static int merge_handshake_packet(gnutls_session_t session,
 		if (hsk->start_offset <
 		    session->internals.handshake_recv_buffer[pos].
 		    start_offset
-		    && hsk->end_offset >=
+		    && hsk->end_offset + 1 >=
 		    session->internals.handshake_recv_buffer[pos].
 		    start_offset) {
 			memcpy(&session->internals.
@@ -1057,8 +1109,16 @@ static int get_last_packet(gnutls_session_t session,
 						      [LAST_ELEMENT]);
 			session->internals.handshake_recv_buffer_size--;
 			return 0;
-		} else
-			goto timeout;
+		} else {
+			/* if we don't have a complete handshake message, but we
+			 * have queued data waiting, try again to reconstruct the
+			 * handshake packet, using the queued */
+			if (recv_buf[LAST_ELEMENT].end_offset != recv_buf[LAST_ELEMENT].length - 1 &&
+			    record_check_unprocessed(session) > 0)
+				return gnutls_assert_val(GNUTLS_E_INT_CHECK_AGAIN);
+			else
+				goto timeout;
+		}
 	} else {		/* TLS */
 
 		if (session->internals.handshake_recv_buffer_size > 0
@@ -1111,7 +1171,7 @@ int _gnutls_parse_record_buffered_msgs(gnutls_session_t session)
 				    gnutls_assert_val
 				    (GNUTLS_E_UNEXPECTED_PACKET);
 
-			/* if we have a half received message the complete it.
+			/* if we have a half received message then complete it.
 			 */
 			remain = recv_buf[0].length -
 			    recv_buf[0].data.length;
@@ -1289,10 +1349,12 @@ _gnutls_handshake_io_recv_int(gnutls_session_t session,
 {
 	int ret;
 	unsigned int tleft = 0;
+	int retries = 7;
 
 	ret = get_last_packet(session, htype, hsk, optional);
-	if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED
-	    && ret != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+	if (ret != GNUTLS_E_AGAIN && ret != GNUTLS_E_INTERRUPTED &&
+	    ret != GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE &&
+	    ret != GNUTLS_E_INT_CHECK_AGAIN) {
 		return gnutls_assert_val(ret);
 	}
 
@@ -1313,24 +1375,33 @@ _gnutls_handshake_io_recv_int(gnutls_session_t session,
 			return gnutls_assert_val(ret);
 	}
 
-	if (htype != (unsigned) -1) {
+	if (htype != (gnutls_handshake_description_t) -1) {
 		ret = handshake_remaining_time(session);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 		tleft = ret;
 	}
 
-	/* if we don't have a complete message waiting for us, try 
-	 * receiving more */
-	ret =
-	    _gnutls_recv_in_buffers(session, GNUTLS_HANDSHAKE, htype,
-				    tleft);
-	if (ret < 0)
-		return gnutls_assert_val_fatal(ret);
+	do {
+		/* if we don't have a complete message waiting for us, try 
+		 * receiving more */
+		ret =
+		    _gnutls_recv_in_buffers(session, GNUTLS_HANDSHAKE, htype,
+					    tleft);
+		if (ret < 0)
+			return gnutls_assert_val_fatal(ret);
 
-	ret = _gnutls_parse_record_buffered_msgs(session);
-	if (ret == 0)
-		ret = get_last_packet(session, htype, hsk, optional);
+		ret = _gnutls_parse_record_buffered_msgs(session);
+		if (ret == 0) {
+			ret = get_last_packet(session, htype, hsk, optional);
+		}
+		/* we put an upper limit (retries) to the number of partial handshake
+		 * messages in a record packet. */
+	} while(IS_DTLS(session) && ret == GNUTLS_E_INT_CHECK_AGAIN && retries-- > 0);
+
+	if (unlikely(IS_DTLS(session) && ret == GNUTLS_E_INT_CHECK_AGAIN)) {
+		ret = gnutls_assert_val(GNUTLS_E_TOO_MANY_HANDSHAKE_PACKETS);
+	}
 
 	return ret;
 }
