@@ -64,6 +64,7 @@ struct find_flags_data_st {
 
 struct find_url_data_st {
 	gnutls_pkcs11_obj_t obj;
+	bool overwrite_exts; /* only valid if looking for a certificate */
 };
 
 struct find_obj_data_st {
@@ -346,6 +347,167 @@ int gnutls_pkcs11_add_provider(const char *name, const char *params)
 }
 
 
+/**
+ * gnutls_pkcs11_obj_set_info:
+ * @obj: should contain a #gnutls_pkcs11_obj_t structure
+ * @itype: Denotes the type of information to be set
+ * @data: the data to set
+ * @data_size: the size of data
+ * @flags: Or sequence of GNUTLS_PKCS11_OBJ_* flags
+ *
+ * This function will set attributes on the provided object.
+ * Available options for @itype are %GNUTLS_PKCS11_OBJ_LABEL,
+ * %GNUTLS_PKCS11_OBJ_ID_HEX, and %GNUTLS_PKCS11_OBJ_ID.
+ *
+ * Returns: %GNUTLS_E_SUCCESS (0) on success or a negative error code on error.
+ *
+ * Since: 3.3.26
+ **/
+int
+gnutls_pkcs11_obj_set_info(gnutls_pkcs11_obj_t obj,
+			   gnutls_pkcs11_obj_info_t itype,
+			   const void *data, size_t data_size,
+			   unsigned flags)
+{
+	struct p11_kit_uri *info = obj->info;
+	struct pkcs11_session_info sinfo;
+	struct ck_attribute *attr;
+	struct ck_attribute a[4];
+	ck_certificate_type_t type = -1;
+	ck_object_class_t class = CKO_CERTIFICATE;
+	ck_object_handle_t pkobj[2];
+	unsigned long count;
+	size_t size;
+	unsigned a_vals;
+	char tmp[128];
+	ck_rv_t rv;
+	int ret;
+
+	PKCS11_CHECK_INIT;
+
+	ret =
+	    pkcs11_open_session(&sinfo, NULL, info,
+				SESSION_WRITE |
+				pkcs11_obj_flags_to_int(flags));
+	if (ret < 0) {
+		gnutls_assert();
+		return ret;
+	}
+
+	/* find the object that matches the URL */
+	a_vals = 0;
+	attr = p11_kit_uri_get_attribute(info, CKA_ID);
+	if (attr) {
+		memcpy(a + a_vals, attr, sizeof(struct ck_attribute));
+		a_vals++;
+	}
+
+	attr = p11_kit_uri_get_attribute(info, CKA_LABEL);
+	if (attr) {
+		memcpy(a + a_vals, attr, sizeof(struct ck_attribute));
+		a_vals++;
+	}
+
+	if (!a_vals) {
+		gnutls_assert();
+		ret = GNUTLS_E_INVALID_REQUEST;
+		goto cleanup;
+	}
+
+	/* Find objects with given class and type */
+	attr = p11_kit_uri_get_attribute(info, CKA_CLASS);
+	if (attr) {
+		if (attr->value
+		    && attr->value_len == sizeof(ck_object_class_t))
+			class = *((ck_object_class_t *) attr->value);
+		if (class == CKO_CERTIFICATE)
+			type = CKC_X_509;
+		memcpy(a + a_vals, attr, sizeof(struct ck_attribute));
+		a_vals++;
+	}
+
+	if (type != (ck_certificate_type_t) - 1) {
+		a[a_vals].type = CKA_CERTIFICATE_TYPE;
+		a[a_vals].value = &type;
+		a[a_vals].value_len = sizeof type;
+		a_vals++;
+	}
+
+	rv = pkcs11_find_objects_init(sinfo.module, sinfo.pks, a,
+				      a_vals);
+	if (rv != CKR_OK) {
+		gnutls_assert();
+		_gnutls_debug_log("p11: FindObjectsInit failed.\n");
+		ret = pkcs11_rv_to_err(rv);
+		goto cleanup;
+	}
+
+	rv = pkcs11_find_objects(sinfo.module, sinfo.pks, pkobj, 2, &count);
+	if (rv != CKR_OK) {
+		gnutls_assert();
+		_gnutls_debug_log("p11: FindObjects failed.\n");
+		ret = pkcs11_rv_to_err(rv);
+		goto cleanup;
+	}
+
+	if (count > 1 || count == 0) {
+		gnutls_assert();
+		if (count > 1)
+			_gnutls_debug_log("p11: More than one objects match (%d)\n", (int)count);
+		ret = GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE;
+		goto cleanup;
+	}
+
+	switch (itype) {
+	case GNUTLS_PKCS11_OBJ_ID_HEX:
+		size = sizeof(tmp);
+		ret = _gnutls_hex2bin(data, data_size, (uint8_t*)tmp, &size);
+		if (ret < 0) {
+			gnutls_assert();
+			goto cleanup;
+		}
+		data = tmp;
+		data_size = size;
+		/* fallthrough */
+	case GNUTLS_PKCS11_OBJ_ID:
+		a[0].type = CKA_ID;
+		a[0].value = (void*)data;
+		a[0].value_len = data_size;
+
+		rv = pkcs11_set_attribute_value(sinfo.module, sinfo.pks, pkobj[0], a, 1);
+		if (rv != CKR_OK) {
+			gnutls_assert();
+			_gnutls_debug_log("p11: set_attribute_value failed.\n");
+			ret = pkcs11_rv_to_err(rv);
+			goto cleanup;
+		}
+
+		break;
+	case GNUTLS_PKCS11_OBJ_LABEL:
+		a[0].type = CKA_LABEL;
+		a[0].value = (void*)data;
+		a[0].value_len = data_size;
+
+		rv = pkcs11_set_attribute_value(sinfo.module, sinfo.pks, pkobj[0], a, 1);
+		if (rv != CKR_OK) {
+			gnutls_assert();
+			_gnutls_debug_log("p11: set_attribute_value failed.\n");
+			ret = pkcs11_rv_to_err(rv);
+			goto cleanup;
+		}
+
+		break;
+	default:
+		gnutls_assert();
+		return GNUTLS_E_INVALID_REQUEST;
+	}
+
+	ret = 0;
+ cleanup:
+ 	pkcs11_close_session(&sinfo);
+ 	return ret;
+}
+
 /**
  * gnutls_pkcs11_obj_get_info:
  * @obj: should contain a #gnutls_pkcs11_obj_t structure
@@ -855,6 +1017,9 @@ gnutls_pkcs11_obj_set_pin_function(gnutls_pkcs11_obj_t obj,
  **/
 void gnutls_pkcs11_obj_deinit(gnutls_pkcs11_obj_t obj)
 {
+	unsigned i;
+	for (i=0;i<obj->pubkey_size;i++)
+		_gnutls_free_datum(&obj->pubkey[i]);
 	_gnutls_free_datum(&obj->raw);
 	p11_kit_uri_free(obj->info);
 	free(obj);
@@ -1307,7 +1472,7 @@ pkcs11_obj_import(ck_object_class_t class, gnutls_pkcs11_obj_t obj,
 
 int pkcs11_read_pubkey(struct ck_function_list *module,
 		       ck_session_handle_t pks, ck_object_handle_t obj,
-		       ck_key_type_t key_type, gnutls_datum_t * pubkey)
+		       ck_key_type_t key_type, gnutls_pkcs11_obj_t pobj)
 {
 	struct ck_attribute a[4];
 	uint8_t *tmp1;
@@ -1339,11 +1504,13 @@ int pkcs11_read_pubkey(struct ck_function_list *module,
 		if (pkcs11_get_attribute_value(module, pks, obj, a, 2) ==
 		    CKR_OK) {
 
-			pubkey[0].data = a[0].value;
-			pubkey[0].size = a[0].value_len;
+			pobj->pubkey[0].data = a[0].value;
+			pobj->pubkey[0].size = a[0].value_len;
 
-			pubkey[1].data = a[1].value;
-			pubkey[1].size = a[1].value_len;
+			pobj->pubkey[1].data = a[1].value;
+			pobj->pubkey[1].size = a[1].value_len;
+
+			pobj->pubkey_size = 2;
 		} else {
 			gnutls_assert();
 			ret = GNUTLS_E_PKCS11_ERROR;
@@ -1361,22 +1528,24 @@ int pkcs11_read_pubkey(struct ck_function_list *module,
 		if ((rv = pkcs11_get_attribute_value(module, pks, obj, a, 2)) ==
 		    CKR_OK) {
 			ret =
-			    _gnutls_set_datum(&pubkey[0], a[0].value,
+			    _gnutls_set_datum(&pobj->pubkey[0], a[0].value,
 					      a[0].value_len);
 
 			if (ret >= 0)
 				ret =
-				    _gnutls_set_datum(&pubkey
+				    _gnutls_set_datum(&pobj->pubkey
 						      [1], a[1].value,
 						      a[1].value_len);
 
 			if (ret < 0) {
 				gnutls_assert();
-				_gnutls_free_datum(&pubkey[1]);
-				_gnutls_free_datum(&pubkey[0]);
+				_gnutls_free_datum(&pobj->pubkey[1]);
+				_gnutls_free_datum(&pobj->pubkey[0]);
 				ret = GNUTLS_E_MEMORY_ERROR;
 				goto cleanup;
 			}
+
+			pobj->pubkey_size = 2;
 		} else {
 			gnutls_assert();
 			ret = pkcs11_rv_to_err(rv);
@@ -1392,12 +1561,13 @@ int pkcs11_read_pubkey(struct ck_function_list *module,
 
 		if ((rv = pkcs11_get_attribute_value(module, pks, obj, a, 2)) ==
 		    CKR_OK) {
-			pubkey[2].data = a[0].value;
-			pubkey[2].size = a[0].value_len;
+			pobj->pubkey[2].data = a[0].value;
+			pobj->pubkey[2].size = a[0].value_len;
 
-			pubkey[3].data = a[1].value;
-			pubkey[3].size = a[1].value_len;
+			pobj->pubkey[3].data = a[1].value;
+			pobj->pubkey[3].size = a[1].value_len;
 
+			pobj->pubkey_size = 4;
 		} else {
 			gnutls_assert();
 			ret = pkcs11_rv_to_err(rv);
@@ -1416,11 +1586,13 @@ int pkcs11_read_pubkey(struct ck_function_list *module,
 		if ((rv = pkcs11_get_attribute_value(module, pks, obj, a, 2)) ==
 		    CKR_OK) {
 
-			pubkey[0].data = a[0].value;
-			pubkey[0].size = a[0].value_len;
+			pobj->pubkey[0].data = a[0].value;
+			pobj->pubkey[0].size = a[0].value_len;
 
-			pubkey[1].data = a[1].value;
-			pubkey[1].size = a[1].value_len;
+			pobj->pubkey[1].data = a[1].value;
+			pobj->pubkey[1].size = a[1].value_len;
+
+			pobj->pubkey_size = 2;
 		} else {
 			gnutls_assert();
 
@@ -1469,7 +1641,7 @@ pkcs11_obj_import_pubkey(struct ck_function_list *module,
 
 		ret =
 		    pkcs11_read_pubkey(module, pks, ctx, key_type,
-				       pobj->pubkey);
+				       pobj);
 		if (ret < 0)
 			return gnutls_assert_val(ret);
 	}
@@ -1677,7 +1849,7 @@ find_obj_url_cb(struct pkcs11_session_info *sinfo,
 	ck_object_class_t class = -1;
 	ck_certificate_type_t type = (ck_certificate_type_t) - 1;
 	ck_rv_t rv;
-	ck_object_handle_t obj;
+	ck_object_handle_t obj = CK_INVALID_HANDLE;
 	unsigned long count, a_vals;
 	int found = 0, ret;
 
@@ -1764,6 +1936,19 @@ find_obj_url_cb(struct pkcs11_session_info *sinfo,
       cleanup:
 	pkcs11_find_objects_final(sinfo);
 
+	if (ret == 0 && find_data->overwrite_exts && find_data->obj->raw.size > 0 && obj != CK_INVALID_HANDLE) {
+		gnutls_datum_t spki;
+		rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_PUBLIC_KEY_INFO, &spki);
+		if (rv == CKR_OK) {
+			ret = pkcs11_override_cert_exts(sinfo, &spki, &find_data->obj->raw);
+			gnutls_free(spki.data);
+			if (ret < 0) {
+				gnutls_assert();
+				return ret;
+			}
+		}
+	}
+
 	return ret;
 }
 
@@ -1816,6 +2001,10 @@ gnutls_pkcs11_obj_import_url(gnutls_pkcs11_obj_t obj, const char *url,
 		return ret;
 	}
 
+	if (flags & GNUTLS_PKCS11_OBJ_FLAG_OVERWRITE_TRUSTMOD_EXT) {
+		find_data.overwrite_exts = 1;
+	}
+
 	ret =
 	    _pkcs11_traverse_tokens(find_obj_url_cb, &find_data, obj->info,
 				    &obj->pin,
@@ -2178,6 +2367,25 @@ retrieve_pin(struct pin_info_st *pin_info, struct p11_kit_uri *info,
 
 	*pin = NULL;
 
+#ifdef P11_KIT_HAS_PIN_VALUE
+	/* First check for pin-value field */
+	pinfile = p11_kit_uri_get_pin_value(info);
+	if (pinfile != NULL) {
+		_gnutls_debug_log("p11: Using pin-value to retrieve PIN\n");
+		*pin = p11_kit_pin_new_for_string(pinfile);
+		if (*pin != NULL)
+			ret = 0;
+	} else { /* try pin-source */
+		/* Check if a pinfile is specified, and use that if possible */
+		pinfile = p11_kit_uri_get_pin_source(info);
+		if (pinfile != NULL) {
+			_gnutls_debug_log("p11: Using pin-source to retrieve PIN\n");
+			ret =
+			    retrieve_pin_from_source(pinfile, token_info, attempts,
+						     user_type, pin);
+		}
+	}
+#else
 	/* Check if a pinfile is specified, and use that if possible */
 	pinfile = p11_kit_uri_get_pinfile(info);
 	if (pinfile != NULL) {
@@ -2186,6 +2394,7 @@ retrieve_pin(struct pin_info_st *pin_info, struct p11_kit_uri *info,
 		    retrieve_pin_from_source(pinfile, token_info, attempts,
 					     user_type, pin);
 	}
+#endif
 
 	/* The global gnutls pin callback */
 	if (ret < 0)
@@ -2768,7 +2977,8 @@ gnutls_pkcs11_obj_list_import_url(gnutls_pkcs11_obj_t * p_list,
 	}
 
 	*n_list = priv.current;
-	memcpy(p_list, priv.p_list, priv.current*sizeof(p_list[0]));
+	if (p_list)
+		memcpy(p_list, priv.p_list, priv.current*sizeof(p_list[0]));
 	gnutls_free(priv.p_list);
 
 	return 0;
@@ -2829,6 +3039,7 @@ gnutls_pkcs11_obj_list_import_url2(gnutls_pkcs11_obj_t ** p_list,
 	if (ret < 0) {
 		gnutls_assert();
 		if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+			*p_list = NULL;
 			*n_list = 0;
 			ret = 0;
 		}
@@ -3212,6 +3423,7 @@ find_cert_cb(struct pkcs11_session_info *sinfo,
 	unsigned tries, i, finalized;
 	ck_bool_t trusted = 1;
 	time_t now;
+	gnutls_datum_t label = {NULL,0}, id = {NULL,0};
 
 	if (info == NULL) {
 		gnutls_assert();
@@ -3318,70 +3530,70 @@ find_cert_cb(struct pkcs11_session_info *sinfo,
 				break;
 			}
 
-			a[0].type = CKA_LABEL;
-			a[0].value = label_tmp;
-			a[0].value_len = sizeof(label_tmp);
-
-			a[1].type = CKA_ID;
-			a[1].value = id_tmp;
-			a[1].value_len = sizeof(id_tmp);
-
 			/* data will contain the certificate */
 			rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_VALUE, &data);
 
-			if (rv == CKR_OK && pkcs11_get_attribute_value
-			    (sinfo->module, sinfo->pks, obj, a,
-			     2) == CKR_OK) {
-				gnutls_datum_t label =
-				    { a[0].value, a[0].value_len };
-				gnutls_datum_t id =
-				    { a[1].value, a[1].value_len };
-
+			if (rv == CKR_OK) {
 				ret = check_found_cert(priv, &data, now);
 				if (ret < 0) {
 					_gnutls_free_datum(&data);
 					continue;
 				}
 
-				if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_OVERWRITE_TRUSTMOD_EXT) {
-					gnutls_datum_t spki;
-					rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_PUBLIC_KEY_INFO, &spki);
-					if (rv == CKR_OK) {
-						ret = pkcs11_override_cert_exts(sinfo, &spki, &data);
-						gnutls_free(spki.data);
-						if (ret < 0) {
-							gnutls_assert();
-							goto cleanup;
-						}
-					}
+				a[0].type = CKA_LABEL;
+				a[0].value = label_tmp;
+				a[0].value_len = sizeof(label_tmp);
+
+				a[1].type = CKA_ID;
+				a[1].value = id_tmp;
+				a[1].value_len = sizeof(id_tmp);
+
+				if (pkcs11_get_attribute_value(sinfo->module, sinfo->pks, obj, a, 2) == CKR_OK) {
+					label.data = a[0].value;
+					label.size = a[0].value_len;
+					id.data = a[1].value;
+					id.size = a[1].value_len;
+
+					found = 1;
+					break;
+				} else {
+					_gnutls_free_datum(&data);
+					_gnutls_debug_log
+					    ("p11: Skipped cert, missing attrs.\n");
 				}
-
-				if (priv->need_import != 0) {
-					ret =
-					    pkcs11_obj_import(class, priv->obj,
-							      &data, &id, &label,
-							      &info->tinfo,
-							      lib_info);
-					if (ret < 0) {
-						gnutls_assert();
-						goto cleanup;
-					}
-				}
-
-
-				found = 1;
-				break;
-			} else {
-				_gnutls_debug_log
-				    ("p11: Skipped cert, missing attrs.\n");
 			}
 		}
 
 		pkcs11_find_objects_final(sinfo);
 		finalized = 1;
 
-		if (found != 0)
+		if (found != 0) {
+			if (priv->flags & GNUTLS_PKCS11_OBJ_FLAG_OVERWRITE_TRUSTMOD_EXT && data.size > 0) {
+				gnutls_datum_t spki;
+				rv = pkcs11_get_attribute_avalue(sinfo->module, sinfo->pks, obj, CKA_PUBLIC_KEY_INFO, &spki);
+				if (rv == CKR_OK) {
+					ret = pkcs11_override_cert_exts(sinfo, &spki, &data);
+					gnutls_free(spki.data);
+					if (ret < 0) {
+						gnutls_assert();
+						goto cleanup;
+					}
+				}
+			}
+
+			if (priv->need_import != 0) {
+				ret =
+				    pkcs11_obj_import(class, priv->obj,
+						      &data, &id, &label,
+						      &info->tinfo,
+						      lib_info);
+				if (ret < 0) {
+					gnutls_assert();
+					goto cleanup;
+				}
+			}
 			break;
+		}
 	}
 
 	if (found == 0) {
@@ -3521,10 +3733,8 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 {
 	int ret;
 	struct find_cert_st priv;
-	uint8_t serial[ASN1_MAX_TL_SIZE+64];
+	uint8_t serial[128];
 	size_t serial_size;
-	uint8_t tag[ASN1_MAX_TL_SIZE];
-	unsigned int tag_size;
 	struct p11_kit_uri *info = NULL;
 
 	PKCS11_CHECK_INIT_RET(0);
@@ -3542,38 +3752,31 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 	}
 
 	/* Attempt searching using the issuer DN + serial number */
-	serial_size = sizeof(serial) - sizeof(tag);
+	serial_size = sizeof(serial);
 	ret =
-	    gnutls_x509_crt_get_serial(cert, serial+sizeof(tag), &serial_size);
+	    gnutls_x509_crt_get_serial(cert, serial, &serial_size);
 	if (ret < 0) {
 		gnutls_assert();
 		ret = 0;
 		goto cleanup;
 	}
 
-	/* PKCS#11 requires a DER encoded serial, wtf. $@(*$@ */
-	tag_size = sizeof(tag);
-	ret = asn1_encode_simple_der(ASN1_ETYPE_INTEGER, serial+sizeof(tag), serial_size,
-		tag, &tag_size);
-	if (ret != ASN1_SUCCESS) {
+	ret = _gnutls_x509_ext_gen_number(serial, serial_size, &priv.serial);
+	if (ret < 0) {
 		gnutls_assert();
 		ret = 0;
 		goto cleanup;
 	}
 
-	memcpy(serial+sizeof(tag)-tag_size, tag, tag_size);
-
-	priv.serial.data = serial+sizeof(tag)-tag_size;
-	priv.serial.size = serial_size + tag_size;
 	priv.crt = cert;
 
 	priv.issuer_dn.data = cert->raw_issuer_dn.data;
 	priv.issuer_dn.size = cert->raw_issuer_dn.size;
 
-	/* when looking for a trusted certificate, we always fully compare
-	 * with the given */
-	if (flags & GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_TRUSTED && !(flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY))
+	/* assume PKCS11_OBJ_FLAG_COMPARE everywhere but DISTRUST info */
+	if (!(flags & GNUTLS_PKCS11_OBJ_FLAG_RETRIEVE_DISTRUSTED) && !(flags & GNUTLS_PKCS11_OBJ_FLAG_COMPARE_KEY)) {
 		flags |= GNUTLS_PKCS11_OBJ_FLAG_COMPARE;
+	}
 
 	priv.flags = flags;
 
@@ -3581,8 +3784,10 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 	    _pkcs11_traverse_tokens(find_cert_cb, &priv, info,
 				    NULL, pkcs11_obj_flags_to_int(flags));
 	if (ret == GNUTLS_E_REQUESTED_DATA_NOT_AVAILABLE) {
+		_gnutls_debug_log("crt_is_known: did not find cert, using issuer DN + serial, using DN only\n");
 		/* attempt searching with the subject DN only */
 		gnutls_assert();
+		gnutls_free(priv.serial.data);
 		memset(&priv, 0, sizeof(priv));
 		priv.crt = cert;
 		priv.flags = flags;
@@ -3595,6 +3800,7 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
 	}
 	if (ret < 0) {
 		gnutls_assert();
+		_gnutls_debug_log("crt_is_known: did not find any cert\n");
 		ret = 0;
 		goto cleanup;
 	}
@@ -3604,6 +3810,7 @@ int gnutls_pkcs11_crt_is_known(const char *url, gnutls_x509_crt_t cert,
       cleanup:
 	if (info)
 		p11_kit_uri_free(info);
+	gnutls_free(priv.serial.data);
 
 	return ret;
 }
